﻿//License: GPLv3 (see LICENSE.txt)
//GameEvents.onShowGUI.Add(something), GameEvents.onHideGUI.Add(something else)

using System.Collections.Generic;
using System.Collections;
using System.Linq;
using System;
using UnityEngine;
using Toolbar;

// Update, Awake, Start

namespace Rotate
{
    [KSPAddon(KSPAddon.Startup.Flight, false)]
    public class VehicleRotator : MonoBehaviour
    {
        //GENERAL VARIABLES
        Dictionary<string, Vector3> lastPosition = new Dictionary<string, Vector3>();
        Dictionary<string, Transform> lastTransform = new Dictionary<string, Transform>();
        Dictionary<string, bool> lastActive = new Dictionary<string, bool>();
        Vessel lastVessel;

        // +++ CFG & DICTIONARYS +++
        Dictionary<string, Vector3> vessel_momentum = new Dictionary<string, Vector3>();
        Dictionary<string, Quaternion> vessel_rotation = new Dictionary<string, Quaternion>();
        Dictionary<string, Vector3> vessel_direction = new Dictionary<string, Vector3>();
        Dictionary<string, CelestialBody> vessel_reference = new Dictionary<string, CelestialBody>();
        String rot_cfg
        {
            get { return KSPUtil.ApplicationRootPath + "/GameData/PersistentRotation/PersistentRotation_" + HighLogic.CurrentGame.Title.TrimEnd("_()SANDBOXCAREERSCIENCE".ToCharArray()).TrimEnd(' ') + ".cfg"; }
        }
        String gui_cfg
        {
            get { return KSPUtil.ApplicationRootPath + "/GameData/PersistentRotation/GUIconfig.cfg"; }
        }
        private void CleanCfg()
        {
            //Debug.Log("Clean in");
            //Clean vessel momentum
            Dictionary<string, Vector3> temp_vessel_momentum = new Dictionary<string, Vector3>();
            foreach (Vessel vessel in FlightGlobals.Vessels)
            {
                temp_vessel_momentum[vessel.id.ToString()] = Vector3.zero;
            }
            foreach (KeyValuePair<string, Vector3> entry in vessel_momentum)
            {
                temp_vessel_momentum[entry.Key] = entry.Value;
            }
            vessel_momentum.Clear();
            foreach (Vessel vessel in FlightGlobals.Vessels)
            {
                vessel_momentum[vessel.id.ToString()] = temp_vessel_momentum[vessel.id.ToString()];
            }
            temp_vessel_momentum.Clear();
            //Clean vessel rotation
            Dictionary<string, Quaternion> temp_vessel_rotation = new Dictionary<string, Quaternion>();
            foreach (Vessel vessel in FlightGlobals.Vessels)
            {
                temp_vessel_rotation[vessel.id.ToString()] = Quaternion.identity;
            }
            foreach (KeyValuePair<string, Quaternion> entry in vessel_rotation)
            {
                temp_vessel_rotation[entry.Key] = entry.Value;
            }
            vessel_rotation.Clear();
            foreach (Vessel vessel in FlightGlobals.Vessels)
            {
                vessel_rotation[vessel.id.ToString()] = temp_vessel_rotation[vessel.id.ToString()];
            }
            temp_vessel_rotation.Clear();
            //Clean vessel direction
            Dictionary<string, Vector3> temp_vessel_direction = new Dictionary<string, Vector3>();
            foreach (Vessel vessel in FlightGlobals.Vessels)
            {
                temp_vessel_direction[vessel.id.ToString()] = Vector3.zero;
            }
            foreach (KeyValuePair<string, Vector3> entry in vessel_direction)
            {
                temp_vessel_direction[entry.Key] = entry.Value;
            }
            vessel_direction.Clear();
            foreach (Vessel vessel in FlightGlobals.Vessels)
            {
                vessel_direction[vessel.id.ToString()] = temp_vessel_direction[vessel.id.ToString()];
            }
            temp_vessel_direction.Clear();
            //Clean vessel reference
            Dictionary<string, CelestialBody> temp_vessel_reference = new Dictionary<string, CelestialBody>();
            foreach (Vessel vessel in FlightGlobals.Vessels)
            {
                temp_vessel_reference[vessel.id.ToString()] = null;
            }
            foreach (KeyValuePair<string, CelestialBody> entry in vessel_reference)
            {
                temp_vessel_reference[entry.Key] = entry.Value;
            }
            vessel_reference.Clear();
            foreach (Vessel vessel in FlightGlobals.Vessels)
            {
                vessel_reference[vessel.id.ToString()] = temp_vessel_reference[vessel.id.ToString()];
            }
            temp_vessel_reference.Clear();
            //Debug.Log("Clean out");
        }
        private void Save()
        {
            //Debug.Log("Save in");
            try
            {
                CleanCfg();

                ConfigNode save = new ConfigNode();
                ConfigNode.CreateConfigFromObject(this, 0, save);

                //Save momentum of all vessels
                ConfigNode momentum = save.AddNode("MOMENTUM");
                foreach (KeyValuePair<string, Vector3> entry in vessel_momentum)
                {
                    momentum.AddValue(entry.Key, KSPUtil.WriteVector(entry.Value));
                }
                //Save rotation of all vessels
                ConfigNode rotation = save.AddNode("ROTATION");
                foreach (KeyValuePair<string, Quaternion> entry in vessel_rotation)
                {
                    rotation.AddValue(entry.Key, KSPUtil.WriteQuaternion(entry.Value));
                }
                //Save direction to the mainBody for each vessel
                ConfigNode direction = save.AddNode("DIRECTION");
                foreach (KeyValuePair<string, Vector3> entry in vessel_direction)
                {
                    direction.AddValue(entry.Key, KSPUtil.WriteVector(entry.Value));
                }
                //Save direction to the mainBody for each vessel
                ConfigNode reference = save.AddNode("REFERENCE");
                foreach (KeyValuePair<string, CelestialBody> entry in vessel_reference)
                {
                    if (entry.Value != null)
                    {
                        reference.AddValue(entry.Key, entry.Value.name);
                    }
                    else
                    {
                        reference.AddValue(entry.Key, "NONE");
                    }
                }
                save.Save(rot_cfg);
            }
            catch (Exception e) { Debug.Log("[PR] Saving not sucessfull: " + e.Message); }

            //Debug.Log("Save out");
        }
        private void Load()
        {
            //Debug.Log("Load in");
            //This is called when all persistent rotation data is being loaded from the cfg file.

            float ran_x;
            float ran_y;
            float ran_z;

            //vessel_momentum from scratch
            vessel_momentum.Clear();
            foreach (Vessel vessel in FlightGlobals.Vessels)
            {
                if (vessel.vesselType == VesselType.SpaceObject || vessel.vesselType == VesselType.Unknown)
                {
                    //Random momentum for every asteroid if it has no momentum already.
                    ran_x = (float)UnityEngine.Random.Range(0, 10) / (float)100;
                    ran_y = (float)UnityEngine.Random.Range(0, 10) / (float)100;
                    ran_z = (float)UnityEngine.Random.Range(0, 10) / (float)100;

                    vessel_momentum[vessel.id.ToString()] = new Vector3(ran_x, ran_y, ran_z);
                }
                else
                {
                    vessel_momentum[vessel.id.ToString()] = Vector3.zero;
                }
            }

            //vessel_rotation from scratch
            vessel_rotation.Clear();
            foreach (Vessel vessel in FlightGlobals.Vessels)
            {
                vessel_rotation[vessel.id.ToString()] = vessel.transform.rotation;
            }

            //vessel_direction from scratch
            vessel_direction.Clear();
            foreach (Vessel vessel in FlightGlobals.Vessels)
            {
                vessel_direction[vessel.id.ToString()] = vessel.mainBody.position - vessel.transform.position;
            }

            //vessel_reference from scratch
            vessel_reference.Clear();
            foreach (Vessel vessel in FlightGlobals.Vessels)
            {
                vessel_reference[vessel.id.ToString()] = null;
            }
            //Open cfg file
            ConfigNode load = ConfigNode.Load(rot_cfg);
            if (load == null)
            {
                Debug.Log("[PR] Save file is empty or not existent!");
                return;
            }
            //Load vessel_momentum
            ConfigNode momentum = load.GetNode("MOMENTUM");
            foreach (ConfigNode.Value s in momentum.values)
            {
                vessel_momentum[s.name] = KSPUtil.ParseVector3(s.value);
            }
            //Load vessel_rotation
            ConfigNode rotation = load.GetNode("ROTATION");
            foreach (ConfigNode.Value s in rotation.values)
            {
                vessel_rotation[s.name] = KSPUtil.ParseQuaternion(s.value);
            }
            //Load vessel_direction
            ConfigNode direction = load.GetNode("DIRECTION");
            foreach (ConfigNode.Value s in direction.values)
            {
                vessel_direction[s.name] = KSPUtil.ParseVector3(s.value);
            }
            //Load vessel_reference
            ConfigNode reference = load.GetNode("REFERENCE");
            foreach (ConfigNode.Value s in reference.values)
            {
                if (s.value != "NONE")
                {
                    foreach (CelestialBody body in FlightGlobals.Bodies)
                    {
                            if (body.name == s.value)
                            {
                                vessel_reference[s.name] = body;
                            }
                    }
                }
                else
                {
                    vessel_reference[s.name] = null;
                }
            }
            //Debug.Log("Load out");
        }

        // +++ GAME EVENTS +++ //
        void OnVesselGoOnRails(Vessel vessel)
        {
            //Debug.Log("VesselOnRails in");
            if (vessel.ActionGroups[KSPActionGroup.SAS])
            {
                //If Autopilot is enabled when vessel goes on rails
                vessel_momentum[vessel.id.ToString()] = Vector3.zero;
            }
            else
            {
                //NOTE: Issue with non-existent ReferenceTransform seems fixed

                if (lastVessel.ReferenceTransform) //If the vessel has a ReferenceTransform part
                {
                    vessel_momentum[vessel.id.ToString()] = Quaternion.Inverse(Quaternion.Inverse(lastVessel.ReferenceTransform.rotation) * lastVessel.transform.rotation) * vessel.angularVelocity;
                }
                else //Kerbals, Flags etc.
                {
                    Debug.LogWarning("No ReferenceTransform found in vessel " + lastVessel.GetName() + ", using fallback method.");
                    vessel_momentum[vessel.id.ToString()] = vessel.angularVelocity;
                }
            }

            vessel_rotation[vessel.id.ToString()] =  vessel.transform.rotation;

            if (vessel_reference.ContainsKey(vessel.id.ToString()))
            {
                if (vessel_reference[vessel.id.ToString()] != null)
                {
                    vessel_direction[vessel.id.ToString()] = vessel_reference[vessel.id.ToString()].position - vessel.transform.position;
                }
            }
            else
            {
                Debug.Log("[PR] No key! 1");
            }
            Save();
            //Debug.Log("VesselOnRails out");
        }
        void OnVesselGoOffRails(Vessel vessel)
        {
            //Debug.Log("VesselOffRails in");
            Load();
            CleanCfg();

            if (vessel.ActionGroups[KSPActionGroup.SAS])
            {
                if (vessel_reference.ContainsKey(vessel.id.ToString()) && vessel_direction.ContainsKey(vessel.id.ToString()) && vessel_rotation.ContainsKey(vessel.id.ToString()))
                {
                    if (vessel_reference[vessel.id.ToString()] != null)
                    {
                        vessel.SetRotation(Quaternion.FromToRotation(vessel_direction[vessel.id.ToString()], vessel_reference[vessel.id.ToString()].position - vessel.transform.position) * vessel_rotation[vessel.id.ToString()]);
                    }
                }
                else
                {
                    Debug.Log("[PR] No key! 2");
                }
            }
            else
            {
                if(vessel_momentum.ContainsKey(vessel.id.ToString()))
                {
                    Vector3 av = vessel_momentum[vessel.id.ToString()];
                    Vector3 COM = vessel.findWorldCenterOfMass();
                    foreach (Part p in vessel.parts) //Applying force on every part
                    {
                        try
                        {
                            if (p.rigidbody == null) continue;
                            p.rigidbody.AddTorque(vessel.transform.rotation * av, ForceMode.VelocityChange);
                            p.rigidbody.AddForce(Vector3.Cross(vessel.transform.rotation * av, (p.rigidbody.position - COM)), ForceMode.VelocityChange);
                        }
                        catch (NullReferenceException nre)
                        {
                            Debug.Log("[PR] NullReferenceException: " + nre.Message);
                        }
                    }
                }
                else
                {
                    Debug.Log("[PR] No key! 3");
                }
            }
            //Debug.Log("VesselOnRails out");
        }
        void Awake()
        {
            //Add GameEvents
            GameEvents.onVesselGoOnRails.Add(OnVesselGoOnRails);
            GameEvents.onVesselGoOffRails.Add(OnVesselGoOffRails);
            GameEvents.onGamePause.Add(OnPause);
            GameEvents.onShowUI.Add(OnShowUI);
            GameEvents.onHideUI.Add(OnHideUI);

            //Initialize GUI Stuff
            mainGuid = Guid.NewGuid().GetHashCode();
            bodyGuid = Guid.NewGuid().GetHashCode();
            configGuid = Guid.NewGuid().GetHashCode();

            MainWindowRect = new Rect(Screen.width / 4, 0, 200, 10); 
            BodyWindowRect = new Rect((Screen.width / 2) - 75, Screen.height / 4, 150, 10);
            ConfigWindowRect = new Rect((Screen.width / 2) - 100, Screen.height / 4, 200, 10);

            LoadGUI();

            if (visibility_mode == 2)
            {
                CreateStockToolbar();
            }
            else if (visibility_mode == 3)
            {
                CreateBlizzyToolbar();
            }
        }
        void Start()
        {
            // Add local functions to Game Events

            Load();
            CleanCfg();

            foreach(Vessel vessel in FlightGlobals.Vessels)
            {
                lastTransform[vessel.id.ToString()] = vessel.ReferenceTransform;
                if (lastTransform.ContainsKey(vessel.id.ToString()) && vessel_reference.ContainsKey(vessel.id.ToString()))
                {
                    if (vessel_reference[vessel.id.ToString()] != null)
                    {
                        lastPosition[vessel.id.ToString()] = (Vector3d)lastTransform[vessel.id.ToString()].position - vessel_reference[vessel.id.ToString()].position;
                    }
                }
                else
                {
                    Debug.Log("[PR] No key! 4");
                }
                
                lastActive[vessel.id.ToString()] = false;

            }
        }
        void OnPause()
        {
            //Debug.Log("OnPause in");
            foreach (Vessel vessel in FlightGlobals.Vessels)
            {
                if (!vessel.packed)
                {
                    vessel_momentum[vessel.id.ToString()] = vessel.angularVelocity;
                    vessel_rotation[vessel.id.ToString()] = vessel.transform.rotation;
                    if (vessel_reference.ContainsKey(vessel.id.ToString()))
                    {
                        if (vessel_reference[vessel.id.ToString()] != null)
                        {
                            vessel_direction[vessel.id.ToString()] = vessel_reference[vessel.id.ToString()].position - vessel.transform.position;
                        }
                    }
                    else
                    {
                        Debug.Log("[PR] No key! 5");
                    }
                }
            }
            Save();
            //Debug.Log("OnPause out");
        }
        void OnDestroy()
        {
            SaveGUI();

            DeleteBlizzyToolbar();
            DeleteStockToolbar();

            //Unbind functions from GameEvents
            GameEvents.onVesselGoOnRails.Remove(OnVesselGoOnRails);
            GameEvents.onVesselGoOffRails.Remove(OnVesselGoOffRails);
            GameEvents.onGamePause.Remove(OnPause);
            GameEvents.onShowUI.Remove(OnShowUI);
            GameEvents.onHideUI.Remove(OnHideUI);
        }

        // +++ PHYSICS +++
        private void SurfaceSAS(Vessel vessel)
        {
            //Debug.Log("SSAS in");
            if(lastActive.ContainsKey(vessel.id.ToString()) && lastTransform.ContainsKey(vessel.id.ToString()) && vessel_reference.ContainsKey(vessel.id.ToString()) && lastPosition.ContainsKey(vessel.id.ToString()))
            {
                if (vessel_reference[vessel.id.ToString()] != null) //Check vessel_reference in current frame
                {
                    if (lastActive[vessel.id.ToString()])
                    {
                        Vector3d newPosition = (Vector3d)lastTransform[vessel.id.ToString()].position - vessel_reference[vessel.id.ToString()].position;
                        QuaternionD delta = FromToRotation(lastPosition[vessel.id.ToString()], newPosition);
                        QuaternionD adjusted = delta * (QuaternionD)vessel.Autopilot.SAS.lockedHeading;
                        vessel.Autopilot.SAS.lockedHeading = adjusted;
                    }
                }
            }
            else
            {
                Debug.Log("[PR] No key! 6");
            }
            //Debug.Log("SSAS out");
        }
        private void RotateWhilePacked(Vessel vessel)
        {
            float ran_x, ran_y, ran_z;
            //Debug.Log("Rotate in");
            Vector3 angvel;
            if (!vessel_momentum.ContainsKey(vessel.id.ToString()))
            {
                 //Random momentum for every asteroid if it has no momentum already.
                 if (vessel.vesselType == VesselType.SpaceObject || vessel.vesselType == VesselType.Unknown)
                 {
                        ran_x = (float)UnityEngine.Random.Range(0, 10) / (float)100;
                        ran_y = (float)UnityEngine.Random.Range(0, 10) / (float)100;
                        ran_z = (float)UnityEngine.Random.Range(0, 10) / (float)100;

                        vessel_momentum[vessel.id.ToString()] = new Vector3(ran_x, ran_y, ran_z);

                        Debug.Log("Random Rotation for Asteroid " + vessel.vesselName + " initiated with magnitude of " + vessel_momentum[vessel.id.ToString()].magnitude);
                }
                else
                {
                    vessel_momentum[vessel.id.ToString()] = Vector3.zero;
                }

                Debug.Log("[PR] No key! 7");
            }

            angvel = vessel_momentum[vessel.id.ToString()];

            if (vessel.situation != Vessel.Situations.LANDED
                && vessel.situation != Vessel.Situations.PRELAUNCH
                && vessel.situation != Vessel.Situations.DOCKED
                && vessel.situation != Vessel.Situations.SPLASHED)
            {
                //Either the vessel continues to rotate, or it follows the planets curvature. This might change.
                if (vessel.ActionGroups[KSPActionGroup.SAS])
                {
                    //TODO: Continue to rotate
                    if (vessel_direction.ContainsKey(vessel.id.ToString()) && vessel_reference.ContainsKey(vessel.id.ToString()) && vessel_rotation.ContainsKey(vessel.id.ToString()))
                    {
                        if (vessel_reference[vessel.id.ToString()] != null)
                        {
                            vessel.SetRotation(Quaternion.FromToRotation(vessel_direction[vessel.id.ToString()], vessel_reference[vessel.id.ToString()].position - vessel.transform.position) * vessel_rotation[vessel.id.ToString()]);
                        }
                    }
                    else
                    {
                        Debug.Log("[PR] No key! 8");
                    }
                }
                else
                {
                    float angle = angvel.magnitude * TimeWarp.CurrentRate;
                    Quaternion rotation = Quaternion.AngleAxis(angle, angvel);
                    vessel.SetRotation(vessel.transform.rotation * rotation);
                }
            }
            //Debug.Log("Rotate out");
        }
        Dictionary<ProtoCrewMember, float> ArtificialGravity = new Dictionary<ProtoCrewMember, float>();
        private void CalculateArtificialGeForce()
        {
            try
            {
                if (FlightGlobals.ActiveVessel.GetCrewCount() > 0)
                {
                    foreach (ProtoCrewMember kerb in FlightGlobals.ActiveVessel.GetVesselCrew())
                    {
                        if (kerb.seat)
                        {
                            float speed = FlightGlobals.ActiveVessel.angularVelocity.magnitude * (kerb.seat.part.rigidbody.position - FlightGlobals.ActiveVessel.findWorldCenterOfMass()).magnitude;
                            float force = (speed * speed) / (kerb.seat.part.rigidbody.position - FlightGlobals.ActiveVessel.findWorldCenterOfMass()).magnitude;
                            ArtificialGravity[kerb] = force;
                            if (ArtificialGravity[kerb] > 10 * 9.81)
                            {
                                ScreenMessages.PostScreenMessage("Danger: Crew Member reaching Artificial Gravity Limit!");
                            }
                            if (force > 9.81 * 22 - 9.81 * kerb.KerbalRef.stupidity * 4)
                            {
                                kerb.KerbalRef.die(); //Let Kerbal die if GeForce exceeds ~20 G
                                ArtificialGravity.Remove(kerb);
                            }
                        }
                    }
                }
            }
            catch (NullReferenceException e) { Debug.Log("[PR] NullReferenceException: " + e.Message); }
        }

        // +++ TIMER +++
        int ticker = 0;
        private void FixedUpdate()
        {   
            if (FlightGlobals.ActiveVessel != lastVessel)
                CleanCfg();

            ticker++;
            if (ticker > 100)
            {
                ticker = 0;
                CleanCfg();
                SaveGUI();
            }

            try
            {
                foreach (Vessel vessel in FlightGlobals.Vessels)
                {
                    if (vessel.packed)
                    {
                        RotateWhilePacked(vessel);
                    }
                    else
                    {
                        SurfaceSAS(vessel);
                    }

                    lastTransform[vessel.id.ToString()] = vessel.ReferenceTransform;
                    if (vessel_reference.ContainsKey(vessel.id.ToString()) && lastTransform.ContainsKey(vessel.id.ToString()))
                    {
                        if (vessel_reference[vessel.id.ToString()] != null)
                        {
                            lastPosition[vessel.id.ToString()] = (Vector3d)lastTransform[vessel.id.ToString()].position - vessel_reference[vessel.id.ToString()].position;
                        }
                        else
                        {
                            lastPosition[vessel.id.ToString()] = Vector3.zero;
                        }
                    }
                    else
                    {
                        lastPosition[vessel.id.ToString()] = Vector3.zero;
                    }
                    if (vessel.ActionGroups[KSPActionGroup.SAS] && !vessel.packed)
                    {
                        lastActive[vessel.id.ToString()] = true;
                    }
                    else
                    {
                        lastActive[vessel.id.ToString()] = false;
                    }
                }
            }
            catch (NullReferenceException nre)
            {
                Debug.Log("NullReference: " + nre.ToString());
            }

            lastVessel = FlightGlobals.ActiveVessel;
        }

        // +++ QUATERNIONS +++
        public QuaternionD FromToRotation(Vector3d fromv, Vector3d tov) //Stock FromToRotation() doesn't work correctly
        {
            Vector3d cross = Vector3d.Cross(fromv, tov);
            double dot = Vector3d.Dot(fromv, tov);
            double wval = dot + Math.Sqrt(fromv.sqrMagnitude * tov.sqrMagnitude);
            double norm = 1.0 / Math.Sqrt(cross.sqrMagnitude + wval * wval);
            return new QuaternionD(cross.x * norm, cross.y * norm, cross.z * norm, wval * norm);
        }

        // +++ GRAPHICAL USER INTERFACE +++
        private IButton button;
        private ApplicationLauncherButton stockButton;
        int mainGuid;
        int bodyGuid;
        int configGuid;
        Rect BodyWindowRect;
        Rect MainWindowRect;
        Rect ConfigWindowRect;
        bool hidden = false; //If F2 is pressed
        bool showMainWindow = true; //Window visible
        bool showBodyWindow = false;
        bool showConfigWindow = false;
        bool MainWindowActive = false; // minimize-maximize
        int visibility_mode = 1; // 1: Always Visible, 2: Stock Toolbar, 3: Blizzys Toolbar

        void OnGUI()
        {
            if (!hidden)
            {
                if (showMainWindow)
                {
                    MainWindowRect = GUILayout.Window(mainGuid, MainWindowRect, MainGUI, "Persistent Rotation");
                }
                if (showBodyWindow)
                {
                    BodyWindowRect = GUILayout.Window(bodyGuid, BodyWindowRect, BodyGUI, "Select Body");
                }
                if (showConfigWindow)
                {
                    ConfigWindowRect = GUILayout.Window(configGuid, ConfigWindowRect, ConfigGUI, "Configure");
                }
            }
        }
        void OnShowUI()
        {
            hidden = false;
        }
        void OnHideUI()
        {
            hidden = true;
        }
        private void MainGUI(int windowID)
        {
            string status;

            GUILayout.BeginVertical();
            if(MainWindowActive)
            {
                status = "minimize";
            }
            else
            {
                status = "maximize";
            }

            if (GUILayout.Button(status, GUILayout.ExpandWidth(true)))
            {
                MainWindowActive = !MainWindowActive;
                MainWindowRect = new Rect(MainWindowRect.x, MainWindowRect.y, 200, 10);
            }
            if (MainWindowActive)
            {
                if (GUILayout.Button("Configure", GUILayout.ExpandWidth(true)))
                {
                    showConfigWindow = !showConfigWindow;
                }
                GUILayout.Space(20);
                GUILayout.Label("BODY RELATIVE ROTATION:");
                if (GUILayout.Button("Select Body", GUILayout.ExpandWidth(true)))
                {
                    showBodyWindow = !showBodyWindow;
                }
                if (vessel_reference.ContainsKey(FlightGlobals.ActiveVessel.id.ToString()))
                {
                    if (vessel_reference[FlightGlobals.ActiveVessel.id.ToString()])
                    {
                        GUILayout.Label("Current Body: " + vessel_reference[FlightGlobals.ActiveVessel.id.ToString()].name);
                    }
                    else
                    {
                        GUILayout.Label("Current Body: none");
                    }
                }
                else
                {
                    GUILayout.Label("Current Body: none");
                }

                GUILayout.Space(20);
                GUILayout.Label("ARTIFICIAL GRAVITY:");


                CalculateArtificialGeForce();
                if (FlightGlobals.ActiveVessel.GetCrewCount() > 0)
                {
                    //TODO: Artificial Gravity!
                    foreach (KeyValuePair<ProtoCrewMember, float> entry in ArtificialGravity)
                    {
                        GUILayout.BeginHorizontal();
                        GUILayout.Label(entry.Key.name);
                        GUILayout.FlexibleSpace();
                        GUILayout.Label(((double)((int)((entry.Value / 9.81) * 10.0)) / 10.0).ToString() + " G");
                        GUILayout.EndHorizontal();
                    }
                }
                else
                {
                    GUILayout.Label("No crew onboard.");
                }
            }
            GUILayout.EndVertical();
            GUI.DragWindow();

        }
        private void BodyGUI(int windowID)
        {
            GUILayout.BeginVertical();
            if (GUILayout.Button("None", GUILayout.ExpandWidth(true)))
            {
                vessel_reference[FlightGlobals.ActiveVessel.id.ToString()] = null;
                showBodyWindow = false;
            }
            GUILayout.Space(20);
            foreach (CelestialBody body in FlightGlobals.Bodies)
            {
                if (GUILayout.Button(body.name, GUILayout.ExpandWidth(true)))
                {
                    vessel_reference[FlightGlobals.ActiveVessel.id.ToString()] = body;
                    showBodyWindow = false;
                }
            }
            GUILayout.EndVertical();
        }
        private void ConfigGUI(int windowID)
        {
            int old_visibility_mode = visibility_mode;

            GUILayout.BeginVertical();
            GUILayout.Space(0);
            GUILayout.Label("Configure Visibility");
            if (GUILayout.Button("Always Visible", GUILayout.ExpandWidth(true)))
            {
                visibility_mode = 1;
                DeleteBlizzyToolbar();
                DeleteStockToolbar();
            }
            if (GUILayout.Button("Stock Toolbar", GUILayout.ExpandWidth(true)))
            {
                visibility_mode = 2;
                CreateStockToolbar();
                DeleteBlizzyToolbar();
            }
            if (ToolbarManager.ToolbarAvailable)
            {
                if (GUILayout.Button("Blizzy's Toolbar", GUILayout.ExpandWidth(true)))
                {
                    visibility_mode = 3;
                    CreateBlizzyToolbar();
                    DeleteStockToolbar();
                }
            }
            GUILayout.EndVertical();
        }
        void CreateBlizzyToolbar()
        {
            if (ToolbarManager.ToolbarAvailable)
            {
                // regular button
                button = ToolbarManager.Instance.add("PersistentRotation", "button");
                button.TexturePath = "PersistentRotation/Textures/texture";
                button.ToolTip = "Toggle PersistentRotation";
                button.OnClick += (e) => 
                {
                    showMainWindow = !showMainWindow;
                    showBodyWindow = showConfigWindow = false;
                };
            }
        }
        void DeleteBlizzyToolbar()
        {
            if (ToolbarManager.ToolbarAvailable)
            {
                if (button != null)
                {
                    button.Destroy();
                }
            }
        }
        void CreateStockToolbar()
        {
            if (!stockButton)
            {
                    stockButton = ApplicationLauncher.Instance.AddModApplication(
                    () =>
                    {
                        showMainWindow = !showMainWindow;
                        showBodyWindow = showConfigWindow = false;
                    },
                    () =>
                    {
                        showMainWindow = !showMainWindow;
                        showBodyWindow = showConfigWindow = false;
                    },
                    null,
                    null,
                    null,
                    null,
                    ApplicationLauncher.AppScenes.FLIGHT,
                    GameDatabase.Instance.GetTexture("PersistentRotation/Textures/texture", false)
                    );
            }
        }
        void DeleteStockToolbar()
        {
            if (stockButton)
            {
                ApplicationLauncher.Instance.RemoveModApplication(stockButton);
                stockButton = null;
            }
        }
        void SaveGUI()
        {
            Debug.LogError("SaveGUI in");
            try
            {
                ConfigNode save = new ConfigNode();
                ConfigNode.CreateConfigFromObject(this, 0, save);
                save.AddValue("x", MainWindowRect.x.ToString());
                save.AddValue("y", MainWindowRect.y.ToString());
                save.AddValue("mode", visibility_mode.ToString());
                save.AddValue("active", MainWindowActive.ToString());
                save.Save(gui_cfg);
            }
            catch (Exception e) { Debug.Log("[PR] Saving not successful: " + e.Message); }

            Debug.LogError("SaveGUI out");
        }
        void LoadGUI()
        {
            ConfigNode load = ConfigNode.Load(gui_cfg);
            if (load == null)
            {
                Debug.Log("[PR] Cfg file is empty or not existent!");
                return;
            }

            foreach(ConfigNode.Value s in load.values)
            {
                if (s.name == "x")
                    MainWindowRect.x = Convert.ToSingle(s.value);

                if (s.name == "y")
                    MainWindowRect.y = Convert.ToSingle(s.value);

                if (s.name == "mode")
                    visibility_mode = Convert.ToInt32(s.value);

                if (s.name == "active")
                    MainWindowActive = Convert.ToBoolean(s.value);
            }
        }

    }
}